from .meta import RunMetaEntryFilter
from .model import ModelFilter
from .optimizationindexset import OptimizationIndexSetFilter
from .optimizationscalar import OptimizationScalarFilter
from .region import RegionFilter
from .run import RunFilter
from .scenario import ScenarioFilter
from .timeseries import TimeSeriesFilter
from .unit import UnitFilter
from .variable import VariableFilter



from typing import Protocol

import pydantic


class PlatformInfo(pydantic.BaseModel):
    name: str
    dsn: str


class Config(Protocol):
    def list_platforms(self) -> list[PlatformInfo]:
        ...

    def get_platform(self, key: str) -> PlatformInfo:
        ...



from typing import ClassVar, Dict

registry: dict = dict()


class ProgrammingError(Exception):
    pass


ExcMeta: type = type(Exception)


class RemoteExceptionMeta(ExcMeta):
    def __new__(cls, name, bases, namespace, **kwargs):
        http_error_name = namespace.get("http_error_name", None)
        if http_error_name is not None:
            try:
                return registry[http_error_name]
            except KeyError:
                registry[http_error_name] = super().__new__(
                    cls, name, bases, namespace, **kwargs
                )
                return registry[http_error_name]
        else:
            raise ProgrammingError("`IxmpError`s must have `http_error_name`.")


class IxmpError(Exception, metaclass=RemoteExceptionMeta):
    _message: str = ""
    http_status_code: int = 500
    http_error_name: ClassVar[str] = "ixmp_error"
    kwargs: Dict

    def __init__(
        self,
        *args,
        message: str | None = None,
        status_code: int | None = None,
        **kwargs,
    ) -> None:
        if len(args) > 0:
            self._message = args[0]
        if message is not None:
            self._message = message
        if status_code is not None:
            self.http_status_code = status_code
        self.kwargs = kwargs
        super().__init__(*args)

    def __str__(self) -> str:
        return self.message

    @property
    def message(self) -> str:
        message = ""
        if self._message != "":
            message = self._message
        elif len(self.kwargs) > 0:
            kwargs_str = ", ".join(
                ["{}={!r}".format(k, v) for k, v in self.kwargs.items()]
            )
            message = kwargs_str

        return message

    @classmethod
    def from_dict(cls, dict_):
        return cls(message=dict_["message"], **dict_["kwargs"])


class InconsistentIamcType(IxmpError):
    http_status_code = 400
    http_error_name = "inconsistent_iamc_type"


class BadRequest(IxmpError):
    http_status_code = 400
    http_error_name = "bad_request"


class ImproperlyConfigured(IxmpError):
    http_error_name = "improperly_configured"


class ManagerApiError(IxmpError):
    http_error_name = "manager_api_error"


class UnknownApiError(IxmpError):
    http_error_name = "unknown_api_error"


class PlatformNotFound(IxmpError):
    http_status_code = 404
    http_error_name = "platform_not_found"


class PlatformNotUnique(IxmpError):
    http_status_code = 409
    http_error_name = "platform_not_unique"


class MissingToken(IxmpError):
    http_status_code = 401
    http_error_name = "missing_token"


class InvalidToken(IxmpError):
    http_status_code = 401
    http_error_name = "invalid_token"


class Forbidden(IxmpError):
    http_status_code = 403
    http_error_name = "forbidden"


class NotFound(IxmpError):
    http_status_code = 404
    http_error_name = "not_found"


class NotUnique(IxmpError):
    http_status_code = 409
    http_error_name = "not_unique"


class DeletionPrevented(IxmpError):
    http_status_code = 400
    http_error_name = "deletion_prevented"


class OperationNotSupported(IxmpError):
    http_status_code = 400
    http_error_name = "operation_not_supported"


class SchemaError(IxmpError):
    http_status_code = 422
    http_error_name = "schema_error"


# == Run ==


class NoDefaultRunVersion(IxmpError):
    http_status_code = 400
    http_error_name = "run_no_default_version"


class InvalidRunMeta(IxmpError):
    http_status_code = 400
    http_error_name = "run_invalid_meta"


# == Filters ==
class BadFilterArguments(IxmpError):
    _message = "The provided filter arguments are malformed."
    http_status_code = 400
    http_error_name = "bad_filter_arguments"

    def __str__(self) -> str:
        return f"{self.kwargs.get('model')} {self.kwargs.get('errors')}"


# == Api Exceptions ==


class InvalidCredentials(IxmpError):
    _message = "The provided credentials are invalid."
    http_status_code = 401
    http_error_name = "invalid_credentials"



from contextlib import suppress

from sqlalchemy import inspect, sql
from sqlalchemy.orm import Mapper
from sqlalchemy.sql import ColumnCollection

from ixmp4.core.exceptions import ProgrammingError


def is_joined(exc: sql.Select, model):
    """Returns `True` if `model` has been joined in `exc`."""
    for visitor in sql.visitors.iterate(exc):
        # Checking for `.join(Child)` clauses
        if visitor.__visit_name__ == "table":
            # Visitor might be of ColumnCollection or so,
            # which cannot be compared to model
            with suppress(TypeError):
                if model == visitor.entity_namespace:  # type: ignore
                    return True
    return False


def get_columns(model_class: type) -> ColumnCollection:
    mapper: Mapper | None = inspect(model_class)
    if mapper is not None:
        return mapper.selectable.columns
    else:
        raise ProgrammingError(f"Model class `{model_class.__name__}` is not mapped.")


def get_pk_columns(model_class: type) -> ColumnCollection:
    columns: ColumnCollection = ColumnCollection()
    for col in get_columns(model_class):
        if col.primary_key:
            columns.add(col)

    return columns.as_readonly()


def get_foreign_columns(model_class: type) -> ColumnCollection:
    columns: ColumnCollection = ColumnCollection()
    for col in get_columns(model_class):
        if len(col.foreign_keys) > 0:
            columns.add(col)

    return columns.as_readonly()



from fastapi import APIRouter, Body, Depends, Query
from pydantic import RootModel

from ixmp4.data import api
from ixmp4.data.backend.base import Backend
from ixmp4.data.db.filters.optimizationscalar import OptimizationScalarFilter

from .. import deps
from ..base import BaseModel
from ..decorators import autodoc

router: APIRouter = APIRouter(
    prefix="/scalars",
    tags=["optimization", "scalars"],
)


class ScalarCreateInput(BaseModel):
    name: str
    value: float
    unit_name: str
    run_id: int


class ScalarUpdateInput(BaseModel):
    value: float | None
    unit_id: int | None


class EnumerationOutput(BaseModel, RootModel):
    root: list[api.Scalar] | api.DataFrame


@autodoc
@router.get("/", response_model=EnumerationOutput)
def enumerate(
    filter: OptimizationScalarFilter = Depends(),
    table: bool | None = Query(False),
    backend: Backend = Depends(deps.get_backend),
):
    return backend.optimization.scalars.enumerate(_filter=filter, table=bool(table))


@autodoc
@router.get("/{id}/", response_model=api.Scalar)
def get_by_id(
    id: int,
    backend: Backend = Depends(deps.get_backend),
):
    return backend.optimization.scalars.get_by_id(id)


@autodoc
@router.patch("/", response_model=EnumerationOutput)
def query(
    filter: OptimizationScalarFilter = Body(
        OptimizationScalarFilter(id=None, name=None, unit__id=None)
    ),
    table: bool = Query(False),
    backend: Backend = Depends(deps.get_backend),
):
    return backend.optimization.scalars.enumerate(_filter=filter, table=bool(table))


@autodoc
@router.patch("/{id}/", response_model=api.Scalar)
def update(
    id: int,
    scalar: ScalarUpdateInput,
    backend: Backend = Depends(deps.get_backend),
):
    return backend.optimization.scalars.update(id, **scalar.model_dump())


@autodoc
@router.post("/", response_model=api.Scalar)
def create(
    scalar: ScalarCreateInput,
    backend: Backend = Depends(deps.get_backend),
):
    return backend.optimization.scalars.create(**scalar.model_dump())



import pandas as pd

from .base import BaseFacade


class MetaRepository(BaseFacade):
    def tabulate(self, **kwargs) -> pd.DataFrame:
        # TODO: accept list of `Run` instances as arg
        # TODO: expand run-id to model-scenario-version-id columns
        return self.backend.meta.tabulate(join_run_index=True, **kwargs).drop(
            columns=["id", "type"]
        )



from typing import ClassVar

from ixmp4.db import filters, utils

from .. import Model, Run


class ModelFilter(filters.BaseFilter, metaclass=filters.FilterMeta):
    id: filters.Id
    name: filters.String

    sqla_model: ClassVar[type] = Model

    def join(self, exc, **kwargs):
        if not utils.is_joined(exc, Model):
            exc = exc.join(Model, Run.model)
        return exc



from ixmp4 import db
from ixmp4.data.db import filters as base
from ixmp4.data.db.iamc.timeseries import TimeSeries
from ixmp4.data.db.run.model import Run
from ixmp4.db import filters, utils

from .model import Scenario


class BaseIamcFilter(filters.BaseFilter, metaclass=filters.FilterMeta):
    def join_datapoints(self, exc: db.sql.Select, session=None):
        if not utils.is_joined(exc, Run):
            exc = exc.join(Run, onclause=Run.scenario__id == Scenario.id)

        if not utils.is_joined(exc, TimeSeries):
            exc = exc.join(TimeSeries, onclause=TimeSeries.run__id == Run.id)

        return exc


class RunFilter(base.RunFilter, metaclass=filters.FilterMeta):
    def join(self, exc, **kwargs):
        if not utils.is_joined(exc, Run):
            exc = exc.join(Run, Run.model)
        return exc

    scenario: base.ScenarioFilter = filters.Field(default=None, exclude=True)


class IamcScenarioFilter(
    base.ScenarioFilter, BaseIamcFilter, metaclass=filters.FilterMeta
):
    region: base.RegionFilter
    variable: base.VariableFilter
    unit: base.UnitFilter
    run: RunFilter = filters.Field(
        default=RunFilter(id=None, version=None, is_default=True)
    )

    def join(self, exc, session=None):
        return super().join_datapoints(exc, session)


class ScenarioFilter(base.ScenarioFilter, BaseIamcFilter, metaclass=filters.FilterMeta):
    iamc: IamcScenarioFilter | filters.Boolean

    def filter_iamc(self, exc, c, v, session=None):
        if v is None:
            return exc

        if v is True:
            return self.join_datapoints(exc, session)
        else:
            ids = self.join_datapoints(db.select(Scenario.id), session)
            exc = exc.where(~Scenario.id.in_(ids))
            return exc

    def join(self, exc, **kwargs):
        return exc



# type: ignore
"""Create optimization_scalar table

Revision ID: 97ba231770e2
Revises: da1fba23f206
Create Date: 2023-11-14 15:59:19.314252

"""
import sqlalchemy as sa
from alembic import op

# Revision identifiers, used by Alembic.
revision = "97ba231770e2"
down_revision = "da1fba23f206"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table(
        "optimization_scalar",
        sa.Column("name", sa.String(length=255), nullable=False),
        sa.Column("value", sa.Float(), nullable=True),
        sa.Column("unit__id", sa.Integer(), nullable=True),
        sa.Column("run__id", sa.Integer(), nullable=False),
        sa.Column("created_at", sa.DateTime(), nullable=True),
        sa.Column("created_by", sa.String(length=255), nullable=True),
        sa.Column(
            "id",
            sa.Integer(),
            sa.Identity(always=False, on_null=True, start=1, increment=1),
            nullable=False,
        ),
        sa.ForeignKeyConstraint(
            ["run__id"],
            ["run.id"],
            name=op.f("fk_optimization_scalar_run__id_run"),
        ),
        sa.ForeignKeyConstraint(
            ["unit__id"],
            ["unit.id"],
            name=op.f("fk_optimization_scalar_unit__id_unit"),
        ),
        sa.PrimaryKeyConstraint("id", name=op.f("pk_optimization_scalar")),
        sa.UniqueConstraint(
            "name", "run__id", name=op.f("uq_optimization_scalar_name_run__id")
        ),
    )
    with op.batch_alter_table("optimization_scalar", schema=None) as batch_op:
        batch_op.create_index(
            batch_op.f("ix_optimization_scalar_run__id"),
            ["run__id"],
            unique=False,
        )
        batch_op.create_index(
            batch_op.f("ix_optimization_scalar_unit__id"),
            ["unit__id"],
            unique=False,
        )

    op.create_table(
        "optimization_scalar_docs",
        sa.Column("description", sa.Text(), nullable=False),
        sa.Column("dimension__id", sa.Integer(), nullable=True),
        sa.Column(
            "id",
            sa.Integer(),
            sa.Identity(always=False, on_null=True, start=1, increment=1),
            nullable=False,
        ),
        sa.ForeignKeyConstraint(
            ["dimension__id"],
            ["optimization_scalar.id"],
            name=op.f("fk_optimization_scalar_docs_dimension__id_optimization_scalar"),
        ),
        sa.PrimaryKeyConstraint("id", name=op.f("pk_optimization_scalar_docs")),
        sa.UniqueConstraint(
            "dimension__id",
            name=op.f("uq_optimization_scalar_docs_dimension__id"),
        ),
    )
    with op.batch_alter_table("optimization_indexset", schema=None) as batch_op:
        batch_op.drop_constraint("uq_optimization_indexset_name", type_="unique")
        batch_op.create_unique_constraint(
            batch_op.f("uq_optimization_indexset_name_run__id"),
            ["name", "run__id"],
        )

    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table("optimization_indexset", schema=None) as batch_op:
        batch_op.drop_constraint(
            batch_op.f("uq_optimization_indexset_name_run__id"), type_="unique"
        )
        batch_op.create_unique_constraint("uq_optimization_indexset_name", ["name"])

    op.drop_table("optimization_scalar_docs")
    with op.batch_alter_table("optimization_scalar", schema=None) as batch_op:
        batch_op.drop_index(batch_op.f("ix_optimization_scalar_unit__id"))
        batch_op.drop_index(batch_op.f("ix_optimization_scalar_run__id"))

    op.drop_table("optimization_scalar")
    # ### end Alembic commands ###



from typing import ClassVar

from ixmp4.data.db.iamc.datapoint import get_datapoint_model
from ixmp4.data.db.iamc.measurand import Measurand
from ixmp4.data.db.iamc.timeseries import TimeSeries
from ixmp4.data.db.iamc.variable import Variable
from ixmp4.data.db.model import Model
from ixmp4.data.db.region import Region
from ixmp4.data.db.run import Run
from ixmp4.data.db.scenario import Scenario
from ixmp4.data.db.unit import Unit
from ixmp4.db import filters, utils


class RegionFilter(filters.BaseFilter, metaclass=filters.FilterMeta):
    name: filters.String
    hierarchy: filters.String

    sqla_model: ClassVar[type] = Region

    def join(self, exc, session):
        model = get_datapoint_model(session)
        if not utils.is_joined(exc, TimeSeries):
            exc = exc.join(TimeSeries, onclause=model.time_series__id == TimeSeries.id)
        if not utils.is_joined(exc, Region):
            exc = exc.join(Region, TimeSeries.region)
        return exc


class UnitFilter(filters.BaseFilter, metaclass=filters.FilterMeta):
    name: filters.String

    sqla_model: ClassVar[type] = Unit

    def join(self, exc, session):
        model = get_datapoint_model(session)
        if not utils.is_joined(exc, TimeSeries):
            exc = exc.join(TimeSeries, onclause=model.time_series__id == TimeSeries.id)
        if not utils.is_joined(exc, Measurand):
            exc = exc.join(Measurand, TimeSeries.measurand)
        if not utils.is_joined(exc, Unit):
            exc = exc.join(Unit, Measurand.unit)

        return exc


class VariableFilter(filters.BaseFilter, metaclass=filters.FilterMeta):
    name: filters.String

    sqla_model: ClassVar[type] = Variable

    def join(self, exc, session):
        model = get_datapoint_model(session)
        if not utils.is_joined(exc, TimeSeries):
            exc = exc.join(TimeSeries, onclause=model.time_series__id == TimeSeries.id)
        if not utils.is_joined(exc, Measurand):
            exc = exc.join(Measurand, TimeSeries.measurand)
        if not utils.is_joined(exc, Variable):
            exc = exc.join(Variable, Measurand.variable)
        return exc


class ModelFilter(filters.BaseFilter, metaclass=filters.FilterMeta):
    name: filters.String

    sqla_model: ClassVar[type] = Model

    def join(self, exc, session):
        model = get_datapoint_model(session)
        if not utils.is_joined(exc, TimeSeries):
            exc = exc.join(TimeSeries, onclause=model.time_series__id == TimeSeries.id)
        if not utils.is_joined(exc, Run):
            exc = exc.join(Run, TimeSeries.run)
        if not utils.is_joined(exc, Model):
            exc = exc.join(Model, Run.model)
        return exc


class ScenarioFilter(filters.BaseFilter, metaclass=filters.FilterMeta):
    name: filters.String

    sqla_model: ClassVar[type] = Scenario

    def join(self, exc, session):
        model = get_datapoint_model(session)
        if not utils.is_joined(exc, TimeSeries):
            exc = exc.join(TimeSeries, onclause=model.time_series__id == TimeSeries.id)
        if not utils.is_joined(exc, Run):
            exc = exc.join(Run, TimeSeries.run)
        if not utils.is_joined(exc, Scenario):
            exc = exc.join(Scenario, onclause=Run.scenario__id == Scenario.id)
        return exc


class RunFilter(filters.BaseFilter, metaclass=filters.FilterMeta):
    id: filters.Id
    default_only: filters.Boolean = filters.Field(True)

    sqla_model: ClassVar[type] = Run

    def join(self, exc, session):
        model = get_datapoint_model(session)
        if not utils.is_joined(exc, TimeSeries):
            exc = exc.join(TimeSeries, model.time_series__id == TimeSeries.id)
        if not utils.is_joined(exc, Run):
            exc = exc.join(Run, TimeSeries.run)
        return exc

    def filter_default_only(self, exc, c, v, **kwargs):
        if v:
            return exc.where(Run.is_default)
        else:
            return exc


class DataPointFilter(filters.BaseFilter, metaclass=filters.FilterMeta):
    """This class is used for filtering data points

    All parameters are optional. Use the field name (or the field alias)
    directly for equality comparisons. For performing an SQL IN operation
    use the field name followed by a double underscore and *in*.

    Parameters
    ----------
    step_year : filters.Integer, Optional
        Filter for data point year, can also be called with "year"
    time_series__id : filters.Id, Optional
        Filter for the id of the time series, can also be called with "time_series_id"
    region : RegionFilter, Optional
        Filter for either region name or hierarchy
    unit : UnitFilter, Optional
        Filter for unit name
    variable : VariableFilter, Optional
        Filter for variable name
    model : ModelFilter, Optional
        Filter for model name
    scenario : ScenarioFilter, Optional
        Filter for the scenario name
    run : RunFilter, Optional
        Filter for the run, options are id or default_only

    Examples
    --------

    Return all data points for a given year.

    >>> DataPointFilter(year = 2020)

    Return all data points a number of years

    >>> DataPointFilter(year__in = [2020, 2025])

    Return all data point that map to a given variable

    >>> DataPointFilter(variable = VariableFilter(name = "variable 1"))

    Note that for actual use providing the filter parameters as key word arguments is
    sufficient. Calling an endpoint could look like this:

    >>> filter = {"variable": {"name": "variable 1"}, "year": 2020}
    >>> iamc.tabulate(**filter)
    """

    step_year: filters.Integer = filters.Field(None, alias="year")
    time_series__id: filters.Id = filters.Field(None, alias="time_series_id")
    region: RegionFilter
    unit: UnitFilter
    variable: VariableFilter
    model: ModelFilter
    scenario: ScenarioFilter
    run: RunFilter = filters.Field(RunFilter())



from fastapi import APIRouter, Body, Depends, Query

from ixmp4.data import api
from ixmp4.data.backend.db import SqlAlchemyBackend as Backend
from ixmp4.data.db.unit.filter import IamcUnitFilter

from .. import deps
from ..base import EnumerationOutput, Pagination

router: APIRouter = APIRouter(
    prefix="/units",
    tags=["iamc", "units"],
)


@router.patch("/", response_model=EnumerationOutput[api.Unit])
def query(
    filter: IamcUnitFilter = Body(IamcUnitFilter()),
    table: bool | None = Query(False),
    pagination: Pagination = Depends(),
    backend: Backend = Depends(deps.get_backend),
):
    return EnumerationOutput(
        results=backend.units.paginate(
            _filter=filter,
            limit=pagination.limit,
            offset=pagination.offset,
            table=bool(table),
        ),
        total=backend.units.count(_filter=filter),
        pagination=pagination,
    )



from typing import ClassVar

from sqlalchemy import UniqueConstraint
from sqlalchemy.orm import validates

from ixmp4 import db
from ixmp4.data import types
from ixmp4.data.abstract import optimization as abstract

from .. import base


class IndexSet(base.BaseModel):
    NotFound: ClassVar = abstract.IndexSet.NotFound
    NotUnique: ClassVar = abstract.IndexSet.NotUnique
    DeletionPrevented: ClassVar = abstract.IndexSet.DeletionPrevented

    name: types.String = db.Column(db.String(255), nullable=False, unique=False)
    elements: types.JsonList = db.Column(db.JsonType, nullable=False, default=[])

    created_at: types.DateTime = db.Column(db.DateTime, nullable=True)
    created_by: types.String = db.Column(db.String(255), nullable=True)

    run__id: types.Integer = db.Column(
        db.Integer, db.ForeignKey("run.id"), nullable=False, index=True
    )

    __table_args__ = (UniqueConstraint(name, run__id),)

    @validates("elements")
    def validate_elements(self, key, value):
        unique = set()
        for element in value:
            if element in unique:
                raise ValueError(f"{element} already defined for IndexSet {self.name}!")
            else:
                unique.add(element)
        return value



from typing import Protocol

import pandas as pd

from ixmp4.data import types

from . import base
from .docs import DocsRepository


class Scenario(base.BaseModel, Protocol):
    """Modeling scenario data model."""

    name: types.String
    "Unique name of the scenario."

    created_at: types.DateTime
    "Creation date/time. TODO"
    created_by: types.String
    "Creator. TODO"

    def __str__(self) -> str:
        return f"<Scenario {self.id} name={self.name}>"


class ScenarioRepository(base.Creator, base.Retriever, base.Enumerator, Protocol):
    docs: DocsRepository

    def create(self, name: str) -> Scenario:
        """Creates a scenario.

        Parameters
        ----------
        name : str
            The name of the scenario.

        Raises
        ------
        :class:`ixmp4.core.exceptions.ScenarioNotUnique`:
            If the scenario with `name` is not unique.

        Returns
        -------
        :class:`ixmp4.data.abstract.Scenario`:
            The created scenario.
        """
        ...

    def get(self, name: str) -> Scenario:
        """Retrieves a scenario.

        Parameters
        ----------
        name : str
            The unique name of the scenario.

        Raises
        ------
        :class:`ixmp4.data.abstract.Scenario.NotFound`:
            If the scenario with `name` does not exist.

        Returns
        -------
        :class:`ixmp4.data.abstract.Scenario`:
            The retrieved scenario.
        """
        ...

    def list(
        self,
        *,
        name: str | None = None,
    ) -> list[Scenario]:
        """Lists scenarios by specified criteria.

        Parameters
        ----------
        name : str
            The name of a scenario. If supplied only one result will be returned.

        Returns
        -------
        Iterable[:class:`ixmp4.data.abstract.Scenario`]:
            List of scenarios.
        """
        ...

    def tabulate(
        self,
        *,
        name: str | None = None,
    ) -> pd.DataFrame:
        """Tabulate scenarios by specified criteria.

        Parameters
        ----------
        name : str
            The name of a scenario. If supplied only one result will be returned.

        Returns
        -------
        :class:`pandas.DataFrame`:
            A data frame with the columns:
                - id
                - name
        """
        ...

    def map(self, *args, **kwargs) -> dict:
        """Return a mapping of scenario-id to scenario-name.

        Returns
        -------
        :class:`dict`
            A dictionary `id` -> `name`
        """
        return dict([(s.id, s.name) for s in self.list(*args, **kwargs)])



from ixmp4.data import abstract

from .. import base
from ..docs import BaseDocsRepository, docs_model
from .model import Model

ModelDocs = docs_model(Model)


class ModelDocsRepository(BaseDocsRepository, base.BaseRepository):
    model_class = ModelDocs
    dimension_model_class = Model

    def list(self, *, dimension_id: int | None = None) -> list[abstract.Docs]:
        return super().list(dimension_id=dimension_id)



import pandas as pd

from ixmp4 import db
from ixmp4.data import abstract
from ixmp4.data.auth.decorators import guard
from ixmp4.data.db.model.model import Model

from .. import base
from .docs import ScenarioDocsRepository
from .model import Scenario


class ScenarioRepository(
    base.Creator[Scenario],
    base.Retriever[Scenario],
    base.Enumerator[Scenario],
    abstract.ScenarioRepository,
):
    model_class = Scenario

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.docs = ScenarioDocsRepository(*args, **kwargs)

        from .filter import ScenarioFilter

        self.filter_class = ScenarioFilter

    def join_auth(self, exc: db.sql.Select) -> db.sql.Select:
        from ixmp4.data.db.run.model import Run

        if not db.utils.is_joined(exc, Run):
            exc = exc.join(Run, onclause=Run.scenario__id == Scenario.id)
        if not db.utils.is_joined(exc, Model):
            exc = exc.join(Model, Run.model)

        return exc

    def add(self, name: str) -> Scenario:
        scenario = Scenario(name=name, **self.get_creation_info())
        self.session.add(scenario)
        return scenario

    @guard("view")
    def get(self, name: str) -> Scenario:
        exc = db.select(Scenario).where(Scenario.name == name)
        try:
            return self.session.execute(exc).scalar_one()
        except db.NoResultFound:
            raise Scenario.NotFound

    @guard("edit")
    def create(self, *args, **kwargs) -> Scenario:
        return super().create(*args, **kwargs)

    @guard("view")
    def list(self, *args, **kwargs) -> list[Scenario]:
        return super().list(*args, **kwargs)

    @guard("view")
    def tabulate(self, *args, **kwargs) -> pd.DataFrame:
        return super().tabulate(*args, **kwargs)



from ixmp4.data.abstract import BaseModel
from ixmp4.data.backend import Backend


class BaseFacade(object):
    backend: Backend

    def __init__(self, _backend: Backend | None = None) -> None:
        if _backend is None:
            raise ValueError(
                f"Cannot initialize `{self.__class__.__name__}` without `_backend`."
            )
        self.backend = _backend


class BaseModelFacade(BaseFacade):
    backend: Backend
    _model: BaseModel

    def __init__(self, *args, _model: BaseModel | None = None, **kwargs) -> None:
        super().__init__(*args, **kwargs)

        if _model is not None:
            self._model = _model



from typing import Protocol

import pandas as pd

from ixmp4.data import types

from . import base
from .docs import DocsRepository


class Region(base.BaseModel, Protocol):
    """Region data model."""

    name: types.String
    "Unique name of the region."
    hierarchy: types.String
    "Region hierarchy."

    created_at: types.DateTime
    "Creation date/time. TODO"
    created_by: types.String
    "Creator. TODO"

    def __str__(self) -> str:
        return f"<Region {self.id} name={self.name}>"


class RegionRepository(
    base.Creator,
    base.Deleter,
    base.Retriever,
    base.Enumerator,
    Protocol,
):
    docs: DocsRepository

    def create(self, name: str, hierarchy: str) -> Region:
        """Creates a region.

        Parameters
        ----------
        name : str
            The name of the region.
        hierarchy : str
            The hierarchy this region is assigned to.

        Raises
        ------
        :class:`ixmp4.data.abstract.Region.NotUnique`:
            If the region with `name` is not unique.


        Returns
        -------
        :class:`ixmp4.ata.base.iamc.Region`:
            The created region.
        """
        ...

    def delete(self, id: int):
        """Deletes a region.

        Parameters
        ----------
        id : int
            The unique integer id of the region.

        Raises
        ------
        :class:`ixmp4.data.abstract.Region.NotFound`:
            If the region with `id` does not exist.
        :class:`ixmp4.data.abstract.Region.DeletionPrevented`:
            If the region with `id` is used in the database, preventing it's deletion.
        """
        ...

    def get(self, name: str) -> Region:
        """Retrieves a region.

        Parameters
        ----------
        name : str
            The unique name of the region.

        Raises
        ------
        :class:`ixmp4.data.abstract.Region.NotFound`:
            If the region with `name` does not exist.

        Returns
        -------
        :class:`ixmp4.data.base.iamc.Region`:
            The retrieved region.
        """
        ...

    def get_or_create(
        self,
        name: str,
        hierarchy: str | None = None,
    ) -> Region:
        try:
            region = self.get(name)
        except Region.NotFound:
            if hierarchy is None:
                raise TypeError(
                    """Argument `hierarchy` is required if `Region` with `name` does not
                     exist."""
                )
            return self.create(name, hierarchy)

        if hierarchy is not None and region.hierarchy != hierarchy:
            raise Region.NotUnique(name)
        else:
            return region

    def list(
        self,
        *,
        name: str | None = None,
        hierarchy: str | None = None,
        **kwargs,
    ) -> list[Region]:
        r"""Lists regions by specified criteria.

        Parameters
        ----------
        name : str
            The name of a region. If supplied only one result will be returned.
        hierarchy : str
            The hierarchy of a region.
        \*\*kwargs: any
            More filter parameters as specified in
            `ixmp4.data.db.region.filters.RegionFilter`.

        Returns
        -------
        Iterable[:class:`ixmp4.data.abstract.Region`]:
            List of regions.
        """
        ...

    def tabulate(
        self,
        *,
        name: str | None = None,
        hierarchy: str | None = None,
        **kwargs,
    ) -> pd.DataFrame:
        r"""Tabulate regions by specified criteria.

        Parameters
        ----------
        name : str
            The name of a region. If supplied only one result will be returned.
        hierarchy : str
            The hierarchy of a region.
        \*\*kwargs: any
            More filter parameters as specified in
            `ixmp4.data.db.region.filters.RegionFilter`.

        Returns
        -------
        :class:`pandas.DataFrame`:
            A data frame with the columns:
                - id
                - name
        """
        ...



from typing import ClassVar

from ixmp4 import db
from ixmp4.data import abstract, types

from .. import base


class Model(base.BaseModel):
    NotFound: ClassVar = abstract.Model.NotFound
    NotUnique: ClassVar = abstract.Model.NotUnique
    DeletionPrevented: ClassVar = abstract.Model.DeletionPrevented

    name: types.String = db.Column(db.String(255), unique=True)

    created_at: types.DateTime = db.Column(db.DateTime, nullable=True)
    created_by: types.String = db.Column(db.String(255), nullable=True)



from typing import Iterable, Protocol

import pandas as pd

from ixmp4.data import types

from .. import base
from ..docs import DocsRepository


class Scalar(base.BaseModel, Protocol):
    """Scalar data model."""

    name: types.String
    """Unique name of the Scalar."""
    value: types.Float
    """Value of the Scalar."""
    unit__id: types.Integer
    "Foreign unique integer id of a unit."
    unit: types.Mapped
    "Associated unit."
    run__id: types.Integer
    "Foreign unique integer id of a run."

    created_at: types.DateTime
    "Creation date/time. TODO"
    created_by: types.String
    "Creator. TODO"

    def __str__(self) -> str:
        return f"<Scalar {self.id} name={self.name}>"


class ScalarRepository(
    base.Creator,
    base.Retriever,
    base.Enumerator,
    Protocol,
):
    docs: DocsRepository

    def create(self, name: str, value: float, unit_name: str, run_id: int) -> Scalar:
        """Creates a Scalar.

        Parameters
        ----------
        name : str
            The name of the Scalar.
        value : float
            The value of the Scalar.
        unit_name : str
            The name of the :class:`ixmp4.data.abstract.Unit` for which this Scalar is
            defined.
        run_id : int
            The id of the :class:`ixmp4.data.abstract.Run` for which this Scalar is
            defined.

        Raises
        ------
        :class:`ixmp4.data.abstract.optimization.Scalar.NotUnique`:
            If the Scalar with `name` already exists for the Run with `run_id`.

        Returns
        -------
        :class:`ixmp4.data.abstract.optimization.Scalar`:
            The created Scalar.
        """
        ...

    def update(
        self, id: int, value: float | None = None, unit_id: int | None = None
    ) -> Scalar:
        """Updates a Scalar.

        Parameters
        ----------
        id : int
            The integer id of the Scalar.
        value : float, optional
            The value of the Scalar.
        unit_id : int, optional
            The id of the :class:`ixmp4.data.abstract.Unit` for which this Scalar is
            defined.

        Returns
        -------
        :class:`ixmp4.data.abstract.optimization.Scalar`:
            The updated Scalar.
        """
        ...

    def get(self, run_id: int, name: str) -> Scalar:
        """Retrieves a Scalar.

        Parameters
        ----------
        run_id : int
            The id of the :class:`ixmp4.data.abstract.Run` for which this Scalar is
            defined.
        name : str
            The name of the Scalar.

        Raises
        ------
        :class:`ixmp4.data.abstract.optimization.Scalar.NotFound`:
            If the Scalar with `name` does not exist.

        Returns
        -------
        :class:`ixmp4.data.abstract.optimization.Scalar`:
            The retrieved Scalar.
        """
        ...

    def get_by_id(self, id: int) -> Scalar:
        """Retrieves a Scalar by it's id.

        Parameters
        ----------
        id : int
            Unique integer id.

        Raises
        ------
        :class:`ixmp4.data.abstract.optimization.Scalar.NotFound`.
            If the Scalar with `id` does not exist.

        Returns
        -------
        :class:`ixmp4.data.abstract.optimization.Scalar`:
            The retrieved Scalar.
        """
        ...

    def list(self, *, name: str | None = None, **kwargs) -> Iterable[Scalar]:
        r"""Lists Scalars by specified criteria.

        Parameters
        ----------
        name : str
            The name of a Scalar. If supplied only one result will be returned.
        # TODO: Update kwargs
        \*\*kwargs: any
            More filter parameters as specified in
            `ixmp4.data.db.iamc.variable.filters.VariableFilter`.

        Returns
        -------
        Iterable[:class:`ixmp4.data.abstract.optimization.Scalar`]:
            List of Scalars.
        """
        ...

    def tabulate(self, *, name: str | None = None, **kwargs) -> pd.DataFrame:
        r"""Tabulate Scalars by specified criteria.

        Parameters
        ----------
        name : str
            The name of a Scalar. If supplied only one result will be returned.
        # TODO: Update kwargs
        \*\*kwargs: any
            More filter parameters as specified in
            `ixmp4.data.db.iamc.variable.filters.VariableFilter`.

        Returns
        -------
        :class:`pandas.DataFrame`:
            A data frame with the columns:
                - id
                - name
                - value
                - `ixmp4.data.abstract.Unit`
                - `ixmp4.data.abstract.Unit`.id
        """
        ...



from ixmp4.data.db import filters as base
from ixmp4.data.db.iamc.datapoint import get_datapoint_model
from ixmp4.data.db.iamc.timeseries import TimeSeries
from ixmp4.data.db.run.model import Run
from ixmp4.db import filters, utils


class IamcRunFilter(filters.BaseFilter, metaclass=filters.FilterMeta):
    region: base.RegionFilter
    variable: base.VariableFilter
    unit: base.UnitFilter

    def join(self, exc, session=None):
        if not utils.is_joined(exc, TimeSeries):
            exc = exc.join(TimeSeries, onclause=TimeSeries.run__id == Run.id)

        model = get_datapoint_model(session)
        if not utils.is_joined(exc, model):
            exc = exc.join(model, onclause=model.time_series__id == TimeSeries.id)
        return exc


class RunFilter(base.RunFilter, metaclass=filters.FilterMeta):
    iamc: IamcRunFilter | filters.Boolean

    def join(self, exc, **kwargs):
        return exc



from typing import Optional

import pandas as pd
import pandera as pa
from pandera.typing import Series

from ixmp4.data.abstract import DataPoint as DataPointModel
from ixmp4.data.abstract import Run
from ixmp4.data.backend import Backend

from ..base import BaseFacade
from ..utils import substitute_type
from .variable import VariableRepository


def to_dimensionless(df: pd.DataFrame) -> pd.DataFrame:
    if "dimensionless" in df.unit:
        raise ValueError(
            "Unit name 'dimensionless' is reserved, use an empty string '' instead."
        )
    df.replace(to_replace={"unit": ""}, value="dimensionless", inplace=True)
    return df


class RemoveDataPointFrameSchema(pa.DataFrameModel):
    type: Optional[Series[pa.String]] = pa.Field(isin=[t for t in DataPointModel.Type])
    step_year: Optional[Series[pa.Int]] = pa.Field(coerce=True, nullable=True)
    step_datetime: Optional[Series[pa.DateTime]] = pa.Field(coerce=True, nullable=True)
    step_category: Optional[Series[pa.String]] = pa.Field(nullable=True)

    region: Optional[Series[pa.String]] = pa.Field(coerce=True)
    unit: Optional[Series[pa.String]] = pa.Field(coerce=True)
    variable: Optional[Series[pa.String]] = pa.Field(coerce=True)


class AddDataPointFrameSchema(RemoveDataPointFrameSchema):
    value: Series[pa.Float] = pa.Field(coerce=True)


MAP_STEP_COLUMN = {
    "ANNUAL": "step_year",
    "CATEGORICAL": "step_year",
    "DATETIME": "step_datetime",
}


def convert_to_std_format(df: pd.DataFrame, join_runs: bool) -> pd.DataFrame:
    df.rename(columns={"step_category": "subannual"}, inplace=True)

    if set(df.type.unique()).issubset(["ANNUAL", "CATEGORICAL"]):
        df.rename(columns={"step_year": "year"}, inplace=True)
        time_col = "year"
    else:

        def map_step_column(df: pd.Series):
            df["time"] = df[MAP_STEP_COLUMN[df.type]]
            return df

        df = df.apply(map_step_column, axis=1)
        time_col = "time"

    columns = ["model", "scenario", "version"] if join_runs else []
    columns += ["region", "variable", "unit"] + [time_col]
    if "subannual" in df.columns:
        columns += ["subannual"]
    return df[columns + ["value"]]


def normalize_df(df: pd.DataFrame, raw: bool, join_runs: bool) -> pd.DataFrame:
    if not df.empty:
        df = df.drop(columns=["time_series__id"])
        df.unit = df.unit.replace({"dimensionless": ""})
        if raw is False:
            return convert_to_std_format(df, join_runs)
    return df


class RunIamcData(BaseFacade):
    """IAMC data.

    Parameters
    ----------
    backend : ixmp4.data.backend.Backend
        Data source backend.
    run : ixmp4.base.run.Run
        Model run.
    """

    run: Run

    def __init__(self, *args, run: Run, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.run = run

    def _contract_parameters(self, df: pd.DataFrame) -> pd.DataFrame:
        ts_df = df[["region", "variable", "unit", "run__id"]].drop_duplicates()
        self.backend.iamc.timeseries.bulk_upsert(ts_df, create_related=True)

        ts_df = self.backend.iamc.timeseries.tabulate(
            join_parameters=True,
            run={"id": self.run.id, "default_only": False},
        )
        ts_df = ts_df.rename(columns={"id": "time_series__id"})

        return pd.merge(
            df,
            ts_df,
            how="left",
            on=["run__id", "region", "unit", "variable"],
            suffixes=(None, "_y"),
        )

    def add(
        self,
        df: pd.DataFrame,
        type: Optional[DataPointModel.Type] = None,
    ):
        df = AddDataPointFrameSchema.validate(df)  # type:ignore
        df = to_dimensionless(df.copy())
        df["run__id"] = self.run.id
        df = self._contract_parameters(df)
        substitute_type(df, type)
        self.backend.iamc.datapoints.bulk_upsert(df)

    def remove(
        self,
        df: pd.DataFrame,
        type: Optional[DataPointModel.Type] = None,
    ):
        df = RemoveDataPointFrameSchema.validate(df)  # type:ignore
        df = to_dimensionless(df.copy())
        df["run__id"] = self.run.id
        df = self._contract_parameters(df)
        substitute_type(df, type)
        df = df.drop(columns=["unit", "variable", "region"])
        self.backend.iamc.datapoints.bulk_delete(df)

    def tabulate(
        self,
        *,
        variable: dict | None = None,
        region: dict | None = None,
        unit: dict | None = None,
        raw: bool = False,
    ) -> pd.DataFrame:
        df = self.backend.iamc.datapoints.tabulate(
            join_parameters=True,
            join_runs=False,
            run={"id": self.run.id, "default_only": False},
            variable=variable,
            region=region,
            unit=unit,
        ).dropna(how="all", axis="columns")

        return normalize_df(df, raw, False)


class PlatformIamcData(BaseFacade):
    variables: VariableRepository

    def __init__(self, _backend: Backend | None = None) -> None:
        self.variables = VariableRepository(_backend=_backend)
        super().__init__(_backend=_backend)

    def tabulate(self, *, join_runs: bool = True, raw: bool = False, **kwargs):
        df = self.backend.iamc.datapoints.tabulate(
            join_parameters=True, join_runs=join_runs, **kwargs
        ).dropna(how="all", axis="columns")

        return normalize_df(df, raw, join_runs)



import logging
from json.decoder import JSONDecodeError
from typing import Any, ClassVar, Generic, Type, TypeVar

import httpx
import pandas as pd
from pydantic import BaseModel as PydanticBaseModel
from pydantic import ConfigDict, Field, GetCoreSchemaHandler
from pydantic_core import CoreSchema, core_schema

from ixmp4.conf import settings
from ixmp4.core.exceptions import (
    ImproperlyConfigured,
    IxmpError,
    UnknownApiError,
    registry,
)

logger = logging.getLogger(__name__)


class BaseModel(PydanticBaseModel):
    NotFound: ClassVar[type[IxmpError]]
    NotUnique: ClassVar[type[IxmpError]]

    model_config = ConfigDict(from_attributes=True)


def df_to_dict(df: pd.DataFrame) -> dict:
    columns = []
    dtypes = []
    for c in df.columns:
        columns.append(c)
        dtypes.append(df[c].dtype.name)

    return {
        "index": df.index.to_list(),
        "columns": columns,
        "dtypes": dtypes,
        "data": df.values.tolist(),
    }


class DataFrame(PydanticBaseModel):
    index: list | None = Field(None)
    columns: list[str] | None
    dtypes: list[str] | None
    data: list | None

    model_config = ConfigDict(json_encoders={pd.Timestamp: lambda x: x.isoformat()})

    @classmethod
    def __get_pydantic_core_schema__(
        cls, source_type: Any, handler: GetCoreSchemaHandler
    ) -> CoreSchema:
        return core_schema.no_info_before_validator_function(cls.validate, handler(cls))
        # yield cls.validate

    @classmethod
    def validate(cls, df: pd.DataFrame | dict):
        if isinstance(df, pd.DataFrame):
            return df_to_dict(df)
        else:
            return df

    def to_pandas(self) -> pd.DataFrame:
        df = pd.DataFrame(
            index=self.index or None,
            columns=self.columns,
            data=self.data,
        )
        if self.columns and self.dtypes:
            for c, dt in zip(self.columns, self.dtypes):
                # there seems to be a type incompatbility between StrDtypeArg and str
                df[c] = df[c].astype(dt)  # type: ignore
        return df


ModelType = TypeVar("ModelType", bound=BaseModel)


class BaseRepository(Generic[ModelType]):
    model_class: Type[ModelType]
    prefix: ClassVar[str]
    enumeration_method: str = "PATCH"

    client: httpx.Client

    def __init__(self, client: httpx.Client, *args, **kwargs) -> None:
        self.client = client

    def sanitize_params(self, params: dict):
        return {k: params[k] for k in params if params[k] is not None}

    def get_remote_exception(self, res: httpx.Response, status_code: int):
        try:
            json = res.json()
        except (ValueError, JSONDecodeError):
            raise UnknownApiError(content=res.text, response_status=res.status_code)

        try:
            error_name = json["error_name"]
        except KeyError:
            raise UnknownApiError(json_data=json, response_status=status_code)

        try:
            exc = registry[error_name]
        except KeyError:
            raise ImproperlyConfigured(
                "Could not find remote exception in registry. "
                "Are you sure client and server ixmp versions are compatible?"
            )
        return exc.from_dict(json)

    def _request(
        self,
        method: str,
        path: str,
        params: dict | None = None,
        json: dict | None = None,
        **kwargs,
    ) -> dict | list | None:
        """Sends a request and reraises remote exception if thrown."""
        if params is None:
            params = {}
        else:
            params = self.sanitize_params(params)

        res = self.client.request(method, path, params=params, json=json, **kwargs)

        if res.status_code >= 400:
            if res.status_code == 413:
                raise ImproperlyConfigured(
                    "Received status code 413 (Payload Too Large). "
                    "Consider decreasing `IXMP4_DEFAULT_UPLOAD_CHUNK_SIZE` "
                    f"(current: {settings.default_upload_chunk_size})."
                )
            raise self.get_remote_exception(res, res.status_code)
        else:
            try:
                return res.json()
            except JSONDecodeError:
                if res.status_code < 300 and res.text == "":
                    return None
            except ValueError:
                pass
            raise UnknownApiError(res.text)

    def _get_by_id(self, id: int, *args, **kwargs) -> dict[str, Any]:
        # we can assume this type on create endpoints
        return self._request("GET", self.prefix + str(id) + "/", **kwargs)  # type: ignore

    def _request_enumeration(
        self,
        table: bool = False,
        params: dict | None = None,
        json: dict | None = None,
    ):
        """Convenience method for requests to the enumeration endpoint."""
        if params is None:
            params = {}

        return self._request(
            self.enumeration_method,
            self.prefix,
            params={**params, "table": table},
            json=json,
        )

    def _handle_pagination(
        self,
        data: dict,
        table: bool = False,
        params: dict | None = None,
        json: dict | None = None,
    ) -> list[list] | list[dict]:
        """Handles paginated response and sends subsequent requests if necessary.
        Returns aggregated pages as a list."""

        total = data.pop("total")
        pagination = data.pop("pagination")
        offset = pagination.pop("offset")
        limit = pagination.pop("limit")
        if total <= offset + limit:
            return [data.pop("results")]
        else:
            new_params = {"limit": limit, "offset": offset + limit}
            if params is not None:
                params.update(new_params)
            else:
                params = new_params
            next_data = self._request_enumeration(table=table, params=params, json=json)
            pages = self._handle_pagination(
                next_data, table=table, params=params, json=json
            )
            return [data.pop("results")] + pages

    def _list(
        self, params: dict | None = None, json: dict | None = None, **kwargs
    ) -> list[ModelType]:
        data = self._request_enumeration(params=params, table=False, json=json)
        if isinstance(data, dict):
            # we can assume this type on list endpoints
            pages: list[list] = self._handle_pagination(
                data, table=False, params=params, json=json
            )  # type: ignore
            results = [i for page in pages for i in page]
        else:
            results = data
        return [self.model_class(**i) for i in results]

    def _tabulate(
        self, params: dict | None = {}, json: dict | None = None, **kwargs
    ) -> pd.DataFrame:
        data = self._request_enumeration(table=True, params=params, json=json)
        pagination = data.get("pagination", None)
        if pagination is not None:
            # we can assume this type on table endpoints
            pages: list[dict] = self._handle_pagination(
                data,
                table=True,
                params=params,
                json=json,
            )  # type: ignore
            dfs = [DataFrame(**page).to_pandas() for page in pages]
            return pd.concat(dfs)
        else:
            return DataFrame(**data).to_pandas()

    def _create(self, *args, **kwargs) -> dict[str, Any]:
        # we can assume this type on create endpoints
        return self._request("POST", *args, **kwargs)  # type: ignore

    def _delete(self, id: int):
        self._request("DELETE", self.prefix + str(id))


class Retriever(BaseRepository[ModelType]):
    def get(self, **kwargs) -> ModelType:
        if self.enumeration_method == "GET":
            list_ = self._list(params=kwargs)
        else:
            list_ = self._list(json=kwargs)

        try:
            [obj] = list_
        except ValueError as e:
            raise self.model_class.NotFound(
                f"Expected exactly one result, got {len(list_)} instead."
            ) from e
        return obj


class Creator(BaseRepository[ModelType]):
    def create(self, **kwargs) -> ModelType:
        res = self._create(
            self.prefix,
            json=kwargs,
        )
        return self.model_class(**res)


class Deleter(BaseRepository[ModelType]):
    def delete(self, id: int):
        self._delete(id)


class Lister(BaseRepository[ModelType]):
    def list(self, *args, **kwargs) -> list[ModelType]:
        return self._list(json=kwargs)


class Tabulator(BaseRepository[ModelType]):
    def tabulate(self, *args, **kwargs) -> pd.DataFrame:
        return self._tabulate(json=kwargs)


class Enumerator(Lister[ModelType], Tabulator[ModelType]):
    def enumerate(
        self, *args, table: bool = False, **kwargs
    ) -> list[ModelType] | pd.DataFrame:
        if table:
            return self.tabulate(*args, **kwargs)
        else:
            return self.list(*args, **kwargs)


class BulkOperator(BaseRepository[ModelType]):
    def yield_chunks(self, df: pd.DataFrame, chunk_size: int):
        for _, chunk in df.groupby(df.index // chunk_size):
            yield chunk


class BulkUpserter(BulkOperator[ModelType]):
    def bulk_upsert(
        self,
        df: pd.DataFrame,
        chunk_size: int = settings.default_upload_chunk_size,
        **kwargs,
    ):
        for chunk in self.yield_chunks(df, chunk_size):
            self.bulk_upsert_chunk(chunk, **kwargs)

    def bulk_upsert_chunk(self, df: pd.DataFrame, **kwargs) -> None:
        dict_ = df_to_dict(df)
        json_ = DataFrame(**dict_).model_dump_json()
        self._request(
            "POST",
            self.prefix + "bulk/",
            params=kwargs,
            content=json_,
        )


class BulkDeleter(BulkOperator[ModelType]):
    def bulk_delete(
        self,
        df: pd.DataFrame,
        chunk_size: int = settings.default_upload_chunk_size,
        **kwargs,
    ):
        for chunk in self.yield_chunks(df, chunk_size):
            self.bulk_delete_chunk(chunk, **kwargs)

    def bulk_delete_chunk(self, df: pd.DataFrame, **kwargs) -> None:
        dict_ = df_to_dict(df)
        json_ = DataFrame(**dict_).model_dump_json()
        self._request(
            "PATCH",
            self.prefix + "bulk/",
            params=kwargs,
            content=json_,
        )



from datetime import datetime, timezone

from fastapi import Depends, FastAPI, Path, Request
from fastapi.encoders import jsonable_encoder
from fastapi.middleware.cors import CORSMiddleware
from fastapi.responses import JSONResponse

from ixmp4.conf import settings
from ixmp4.core.exceptions import IxmpError

from . import deps, docs, meta, model, region, run, scenario, unit
from .base import BaseModel
from .iamc import datapoint, timeseries
from .iamc import model as iamc_model
from .iamc import region as iamc_region
from .iamc import scenario as iamc_scenario
from .iamc import unit as iamc_unit
from .iamc import variable as iamc_variable
from .middleware import RequestSizeLoggerMiddleware, RequestTimeLoggerMiddleware
from .optimization import indexset, scalar

v1 = FastAPI(servers=[{"url": "/v1", "description": "v1"}])

if settings.mode == "debug":
    v1.add_middleware(RequestSizeLoggerMiddleware)
    v1.add_middleware(RequestTimeLoggerMiddleware)

v1.add_middleware(
    CORSMiddleware,
    allow_origins=["*"],
    allow_credentials=True,
    allow_methods=["*"],
    allow_headers=["*"],
)

v1.include_router(datapoint.router, prefix="/iamc")
v1.include_router(docs.router)
v1.include_router(iamc_model.router, prefix="/iamc")
v1.include_router(iamc_scenario.router, prefix="/iamc")
v1.include_router(iamc_region.router, prefix="/iamc")
v1.include_router(iamc_unit.router, prefix="/iamc")
v1.include_router(iamc_variable.router, prefix="/iamc")
v1.include_router(indexset.router, prefix="/optimization")
v1.include_router(meta.router)
v1.include_router(model.router)
v1.include_router(region.router)
v1.include_router(run.router)
v1.include_router(scalar.router, prefix="/optimization")
v1.include_router(scenario.router)
v1.include_router(timeseries.router, prefix="/iamc")
v1.include_router(unit.router)


class APIInfo(BaseModel):
    name: str
    version: str
    is_managed: bool
    manager_url: None | str
    utcnow: datetime


@v1.get("/", response_model=APIInfo)
def root(
    platform: str = Path(),
    version: str = Depends(deps.get_version),
):
    return APIInfo(
        name=platform,
        version=version,
        is_managed=settings.managed,
        manager_url=str(settings.manager_url),
        utcnow=datetime.now(tz=timezone.utc),
    )


@v1.exception_handler(IxmpError)
async def http_exception_handler(request: Request, exc: IxmpError):
    return JSONResponse(
        content=jsonable_encoder(
            {
                "message": exc._message,
                "kwargs": exc.kwargs,
                "error_name": exc.http_error_name,
            }
        ),
        status_code=exc.http_status_code,
    )



from typing import ClassVar, Protocol, _ProtocolMeta

import pandas as pd

from ixmp4.core.exceptions import DeletionPrevented, IxmpError, NotFound, NotUnique
from ixmp4.data import types


class BaseMeta(_ProtocolMeta):
    def __init__(self, name, bases, namespace):
        super().__init__(name, bases, namespace)
        self.NotUnique = type(
            self.__name__ + "NotUnique",
            (NotUnique,),
            {"http_error_name": name.lower() + "_not_unique"},
        )
        self.NotFound = type(
            self.__name__ + "NotFound",
            (NotFound,),
            {"http_error_name": name.lower() + "_not_found"},
        )

        self.DeletionPrevented = type(
            self.__name__ + "DeletionPrevented",
            (DeletionPrevented,),
            {"http_error_name": name.lower() + "_deletion_prevented"},
        )


class BaseModel(Protocol, metaclass=BaseMeta):
    NotUnique: ClassVar[type[IxmpError]]
    NotFound: ClassVar[type[IxmpError]]
    DeletionPrevented: ClassVar[type[IxmpError]]

    id: types.Integer
    "Unique integer id."


class BaseRepository(Protocol):
    def __init__(self, *args, **kwargs) -> None:
        ...


class Retriever(BaseRepository, Protocol):
    def get(self, *args, **kwargs) -> BaseModel:
        ...


class Creator(BaseRepository, Protocol):
    def create(self, *args, **kwargs) -> BaseModel:
        ...


class Deleter(BaseRepository, Protocol):
    def delete(self, *args, **kwargs) -> None:
        ...


class Lister(BaseRepository, Protocol):
    def list(self, *args, **kwargs) -> list:
        ...


class Tabulator(BaseRepository, Protocol):
    def tabulate(self, *args, **kwargs) -> pd.DataFrame:
        ...


class Enumerator(Lister, Tabulator, Protocol):
    def enumerate(self, *args, table: bool = False, **kwargs) -> list | pd.DataFrame:
        ...


class BulkUpserter(BaseRepository, Protocol):
    def bulk_upsert(self, *args, **kwargs) -> None:
        ...


class BulkDeleter(BaseRepository, Protocol):
    def bulk_delete(self, *args, **kwargs) -> None:
        ...



from typing import ClassVar

from ixmp4 import db
from ixmp4.core.exceptions import InvalidRunMeta
from ixmp4.data import abstract, types

from .. import base


class RunMetaEntry(base.BaseModel):
    NotFound: ClassVar = abstract.RunMetaEntry.NotFound
    NotUnique: ClassVar = abstract.RunMetaEntry.NotUnique
    DeletionPrevented: ClassVar = abstract.RunMetaEntry.DeletionPrevented

    Type: ClassVar = abstract.RunMetaEntry.Type

    _column_map = {
        abstract.RunMetaEntry.Type.INT: "value_int",
        abstract.RunMetaEntry.Type.STR: "value_str",
        abstract.RunMetaEntry.Type.FLOAT: "value_float",
        abstract.RunMetaEntry.Type.BOOL: "value_bool",
    }

    __table_args__ = (
        db.UniqueConstraint(
            "run__id",
            "key",
        ),
    )
    updateable_columns = [
        "type",
        "value_int",
        "value_str",
        "value_float",
        "value_bool",
    ]

    run__id: types.Integer = db.Column(
        db.Integer,
        db.ForeignKey("run.id"),
        nullable=False,
        index=True,
    )
    run = db.relationship(
        "Run",
        backref="meta",
        foreign_keys=[run__id],
    )

    key: types.String = db.Column(db.String(1023), nullable=False)
    type: types.String = db.Column(db.String(20), nullable=False)

    value_int: types.Integer = db.Column(db.Integer, nullable=True)
    value_str: types.String = db.Column(db.String(1023), nullable=True)
    value_float: types.Float = db.Column(db.Float, nullable=True)
    value_bool: types.Boolean = db.Column(db.Boolean, nullable=True)

    @property
    def value(self) -> abstract.MetaValue:
        type_ = RunMetaEntry.Type(self.type)
        col = self._column_map[type_]
        return getattr(self, col)

    def __init__(self, *args, **kwargs) -> None:
        value = kwargs.pop("value")
        value_type = type(value)
        try:
            type_ = RunMetaEntry.Type.from_pytype(value_type)
            col = self._column_map[type_]
        except KeyError:
            raise InvalidRunMeta(
                f"Invalid type `{value_type}` for value of `RunMetaEntry`."
            )
        kwargs["type"] = type_
        kwargs[col] = value
        super().__init__(*args, **kwargs)



import pandas as pd
from sqlalchemy.exc import NoResultFound

from ixmp4 import db
from ixmp4.data import abstract
from ixmp4.data.auth.decorators import guard

from .. import base
from .docs import UnitDocsRepository
from .model import Unit


class UnitRepository(
    base.Creator[Unit],
    base.Deleter[Unit],
    base.Retriever[Unit],
    base.Enumerator[Unit],
    abstract.UnitRepository,
):
    model_class = Unit

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

        from .filter import UnitFilter

        self.filter_class = UnitFilter
        self.docs = UnitDocsRepository(*args, **kwargs)

    def add(self, name: str) -> Unit:
        unit = Unit(name=name, **self.get_creation_info())
        self.session.add(unit)
        return unit

    @guard("manage")
    def create(self, *args, **kwargs) -> Unit:
        return super().create(*args, **kwargs)

    @guard("manage")
    def delete(self, *args, **kwargs):
        return super().delete(*args, **kwargs)

    @guard("view")
    def get(self, name: str) -> Unit:
        exc: db.sql.Select = db.select(Unit).where(Unit.name == name)
        try:
            return self.session.execute(exc).scalar_one()
        except NoResultFound:
            raise Unit.NotFound

    @guard("view")
    def get_by_id(self, id: int) -> Unit:
        obj = self.session.get(self.model_class, id)

        if obj is None:
            raise Unit.NotFound(id=id)

        return obj

    @guard("view")
    def list(self, *args, **kwargs) -> list[Unit]:
        return super().list(*args, **kwargs)

    @guard("view")
    def tabulate(self, *args, **kwargs) -> pd.DataFrame:
        return super().tabulate(*args, **kwargs)



from typing import ClassVar

from sqlalchemy import UniqueConstraint

from ixmp4 import db
from ixmp4.data import types
from ixmp4.data.abstract import optimization as abstract
from ixmp4.data.db.unit import Unit

from .. import base


class Scalar(base.BaseModel):
    NotFound: ClassVar = abstract.Scalar.NotFound
    NotUnique: ClassVar = abstract.Scalar.NotUnique
    DeletionPrevented: ClassVar = abstract.Scalar.DeletionPrevented

    name: types.String = db.Column(db.String(255), nullable=False, unique=False)
    value: types.Float = db.Column(db.Float, nullable=True, unique=False)

    unit: types.Mapped[Unit | None] = db.relationship()
    unit__id: types.Mapped[int | None] = db.Column(
        db.Integer, db.ForeignKey("unit.id"), index=True
    )

    run__id: types.Integer = db.Column(
        db.Integer, db.ForeignKey("run.id"), nullable=False, index=True
    )

    __table_args__ = (UniqueConstraint(name, run__id),)

    created_at: types.DateTime = db.Column(db.DateTime, nullable=True)
    created_by: types.String = db.Column(db.String(255), nullable=True)



import random
import sys
from datetime import datetime, timedelta
from itertools import cycle
from typing import Generator

import numpy as np
import pandas as pd

from ixmp4.core.platform import Backend, Platform
from ixmp4.core.region import Region
from ixmp4.core.run import Run
from ixmp4.core.unit import Unit
from ixmp4.data.abstract import DataPoint


class MockDataGenerator(object):
    platform: Platform
    backend: Backend

    def __init__(
        self,
        platform: Platform,
        num_models: int,
        num_runs: int,
        num_regions: int,
        num_variables: int,
        num_units: int,
        num_datapoints: int,
    ) -> None:
        self.platform = platform
        self.backend = platform.backend
        self.num_models = num_models
        self.num_runs = num_runs
        self.num_regions = num_regions
        self.num_variables = num_variables
        self.num_units = num_units
        self.num_datapoints = num_datapoints

    def yield_model_names(self):
        for i in range(self.num_models):
            yield f"Model {i}"

    def yield_runs(self, model_names: Generator[str, None, None]):
        scen_per_model = self.num_runs // self.num_models
        if scen_per_model == 0:
            scen_per_model = 1

        scenario_index = 0
        model_name = next(model_names)
        for i in range(self.num_runs):
            run = self.platform.runs.create(model_name, f"Scenario {scenario_index}")
            yield run

            scenario_index += 1
            if scenario_index == scen_per_model:
                run.set_as_default()
                model_name = next(model_names)
                scenario_index = 0

    def yield_regions(self):
        for i in range(self.num_regions):
            name = f"Region {i}"
            try:
                yield self.platform.regions.create(name, "default")
            except Region.NotUnique:
                yield self.platform.regions.get(name)

    def yield_units(self):
        for i in range(self.num_units):
            name = f"Unit {i}"
            try:
                yield self.platform.units.create(name)
            except Unit.NotUnique:
                yield self.platform.units.get(name)

    def yield_variable_names(self):
        for i in range(self.num_variables):
            yield f"Variable {i}"

    def yield_datapoints(
        self,
        runs: Generator[Run, None, None],
        variable_names: Generator[str, None, None],
        units: Generator[Unit, None, None],
        regions: Generator[Region, None, None],
    ):
        dp_count = 0
        for run in runs:
            region_name = next(regions).name
            variable_name = next(variable_names)
            unit_name = next(units).name
            dp_type = random.choice(
                [
                    DataPoint.Type.ANNUAL,
                    DataPoint.Type.CATEGORICAL,
                    DataPoint.Type.DATETIME,
                ]
            )
            df = self.get_datapoints(
                dp_type,
                max=self.num_datapoints - dp_count,
            )
            df["region"] = region_name
            df["variable"] = variable_name
            df["unit"] = unit_name
            run.iamc.add(df, type=dp_type)
            yield df
            dp_count += len(df)
            if self.num_datapoints == dp_count:
                break

    def get_datapoints(self, type: DataPoint.Type, max: int = sys.maxsize):
        df = pd.DataFrame(
            columns=[
                "region",
                "variable",
                "unit",
                "value",
            ],
        )
        if type == DataPoint.Type.ANNUAL:
            amount = min(20, max)
            start_year = random.randint(1950, 2000)
            steps_annual = [start_year + i for i in range(amount)]
            df["step_year"] = steps_annual
        if type == DataPoint.Type.CATEGORICAL:
            amount = min(50, max)
            num_categories = random.randint(2, 10)
            start_year = random.randint(1950, 2000)
            categories = cycle([f"Category {i}" for i in range(num_categories)])
            steps_year = []
            steps_category = []
            for i in range(amount):
                steps_category.append(next(categories))
                steps_year.append(start_year + (i // num_categories))

            df["step_year"] = steps_year
            df["step_category"] = steps_category
        if type == DataPoint.Type.DATETIME:
            amount = min(100, max)
            dt = timedelta(minutes=random.randint(1, 360) * 10)
            start_dt = datetime(
                year=random.randint(1950, 2000),
                month=1,
                day=1,
            ) + timedelta(days=random.randint(0, 365))
            steps_datetime = [start_dt + (dt * i) for i in range(amount)]
            df["step_datetime"] = steps_datetime

        denom = random.randint(10, 100)
        values = np.sin([(i / denom) for i in range(amount)]) * random.random() * 10
        df["value"] = values
        return df

    def generate(self):
        model_names = cycle([n for n in self.yield_model_names()])
        runs = cycle([r for r in self.yield_runs(model_names=model_names)])
        regions = cycle([r for r in self.yield_regions()])
        units = cycle([u for u in self.yield_units()])
        variable_names = cycle([v for v in self.yield_variable_names()])
        for df in self.yield_datapoints(runs, variable_names, units, regions):
            pass



from typing import ClassVar, Type

from ixmp4.data import abstract

from . import base


class Docs(base.BaseModel):
    NotFound: ClassVar = abstract.Docs.NotFound
    NotUnique: ClassVar = abstract.Docs.NotUnique
    DeletionPrevented: ClassVar = abstract.Docs.DeletionPrevented

    id: int
    description: str
    dimension__id: int


class DocsRepository(
    base.Creator[Docs],
    base.Retriever[Docs],
    base.Deleter[Docs],
    base.Enumerator[Docs],
    abstract.DocsRepository,
):
    model_class: Type[Docs]
    enumeration_method = "GET"

    def get(self, dimension_id: int) -> Docs:
        return super().get(dimension_id=dimension_id)

    def set(self, dimension_id: int, description: str) -> Docs:
        res = self._create(
            self.prefix,
            json={
                "dimension_id": dimension_id,
                "description": description,
            },
        )
        return Docs(**res)

    def list(self, *, dimension_id: int | None = None) -> list[Docs]:
        return super()._list(params={"dimension_id": dimension_id})

    def delete(self, dimension_id: int) -> None:
        super().delete(dimension_id)



# flake8: noqa
import importlib.metadata

from ixmp4.core import Model as Model
from ixmp4.core import Platform as Platform
from ixmp4.core import Region as Region
from ixmp4.core import Run as Run
from ixmp4.core import Scenario as Scenario
from ixmp4.core import Unit as Unit
from ixmp4.core import Variable as Variable
from ixmp4.core import IndexSet as IndexSet
from ixmp4.core import Scalar as Scalar
from ixmp4.core.exceptions import InconsistentIamcType as InconsistentIamcType
from ixmp4.core.exceptions import IxmpError as IxmpError
from ixmp4.core.exceptions import NotFound as NotFound
from ixmp4.core.exceptions import NotUnique as NotUnique
from ixmp4.data.abstract import DataPoint as DataPoint

__version__ = "0.7.3"
__version_tuple__ = (0, 7, 3)



from typing import Union

import pandas as pd
import pandera as pa
from pandera.typing import DataFrame, Series
from sqlalchemy.exc import NoResultFound

from ixmp4 import db
from ixmp4.core.decorators import check_types
from ixmp4.data import abstract
from ixmp4.data.auth.decorators import guard
from ixmp4.data.db.model import Model
from ixmp4.data.db.run import Run
from ixmp4.data.db.run.repository import select_joined_run_index

from .. import base
from .model import RunMetaEntry


class RemoveRunMetaEntryFrameSchema(pa.DataFrameModel):
    key: Series[pa.String] = pa.Field(coerce=True)
    run__id: Series[pa.Int] = pa.Field(coerce=True)

    class Config:
        strict = True
        coerce = True


class AddRunMetaEntryFrameSchema(RemoveRunMetaEntryFrameSchema):
    value: Series[pa.Object] = pa.Field(coerce=True)


class UpdateRunMetaEntryFrameSchema(AddRunMetaEntryFrameSchema):
    id: Series[pa.Int] = pa.Field(coerce=True)


class RunMetaEntryRepository(
    base.Creator[RunMetaEntry],
    base.Enumerator[RunMetaEntry],
    base.BulkUpserter[RunMetaEntry],
    base.BulkDeleter[RunMetaEntry],
    abstract.RunMetaEntryRepository,
):
    model_class = RunMetaEntry

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        from .filter import RunMetaEntryFilter

        self.filter_class = RunMetaEntryFilter

    def add(
        self, run__id: int, key: str, value: Union[str, int, bool, float]
    ) -> RunMetaEntry:
        if self.backend.auth_context is not None:
            self.backend.runs.check_access(
                {run__id},
                access_type="edit",
                is_default=None,
                default_only=False,
            )

        entry = RunMetaEntry(run__id=run__id, key=key, value=value)
        self.session.add(entry)
        return entry

    def check_df_access(self, df: pd.DataFrame):
        if self.backend.auth_context is not None:
            ts_ids = set(df["run__id"].unique().tolist())
            self.backend.runs.check_access(
                ts_ids,
                access_type="edit",
                is_default=None,
                default_only=False,
            )

    @guard("edit")
    def create(self, *args, **kwargs) -> RunMetaEntry:
        return super().create(*args, **kwargs)

    @guard("view")
    def get(self, run__id: int, key: str) -> RunMetaEntry:
        if self.backend.auth_context is not None:
            self.backend.runs.check_access(
                {run__id},
                access_type="view",
                is_default=None,
                default_only=False,
            )

        exc = self.select(
            run_id=run__id,
            key=key,
        )

        try:
            return self.session.execute(exc).scalar_one()
        except NoResultFound:
            raise RunMetaEntry.NotFound(
                run__id=run__id,
                key=key,
            )

    @guard("edit")
    def delete(self, id: int) -> None:
        if self.backend.auth_context is not None:
            try:
                pre_exc = db.select(RunMetaEntry).where(RunMetaEntry.id == id)
                meta = self.session.execute(pre_exc).scalar_one()
            except NoResultFound:
                raise RunMetaEntry.NotFound(
                    id=id,
                )
            self.backend.runs.check_access(
                {meta.run__id},
                access_type="edit",
                is_default=None,
                default_only=False,
            )

        exc = db.delete(RunMetaEntry).where(RunMetaEntry.id == id)

        try:
            self.session.execute(exc)
            self.session.commit()
        except NoResultFound:
            raise RunMetaEntry.NotFound(
                id=id,
            )

    def join_auth(self, exc: db.sql.Select) -> db.sql.Select:
        if not db.utils.is_joined(exc, Run):
            exc = exc.join(Run, RunMetaEntry.run)
        if not db.utils.is_joined(exc, Model):
            exc = exc.join(Model, Run.model)

        return super().join_auth(exc)

    @guard("view")
    def list(self, *args, **kwargs) -> list[RunMetaEntry]:
        return super().list(*args, **kwargs)

    @guard("view")
    def tabulate(
        self,
        *args,
        join_run_index: bool = False,
        _raw: bool | None = False,
        **kwargs,
    ) -> pd.DataFrame:
        if _raw:
            return super().tabulate(*args, **kwargs)

        if join_run_index:
            index_columns = ["model", "scenario", "version"]
            _exc = select_joined_run_index(self, **kwargs)
            df = super().tabulate(*args, _exc=_exc, **kwargs)
            df.drop(columns="run__id", inplace=True)
        else:
            index_columns = ["run__id"]
            df = super().tabulate(*args, **kwargs)

        if df.empty:
            return pd.DataFrame(
                [], columns=index_columns + ["id", "type", "key", "value"]
            )

        def map_value_column(df: pd.DataFrame):
            type_str = df.name
            type_ = RunMetaEntry.Type(type_str)
            col = RunMetaEntry._column_map[type_]
            df["value"] = df[col]
            return df.drop(columns=RunMetaEntry._column_map.values())

        return df.groupby("type", group_keys=False).apply(map_value_column)

    @check_types
    @guard("edit")
    def bulk_upsert(self, df: DataFrame[AddRunMetaEntryFrameSchema]) -> None:
        self.check_df_access(df)
        df["type"] = df["value"].map(type).map(RunMetaEntry.Type.from_pytype)

        type_: RunMetaEntry.Type
        for type_, type_df in df.groupby("type"):
            col = RunMetaEntry._column_map[type_]
            null_cols = set(RunMetaEntry._column_map.values()) - set([col])
            type_df["type"] = type_df["type"].map(lambda x: x.value)
            type_df = type_df.rename(columns={"value": col})

            # ensure all other columns are overwritten
            for nc in null_cols:
                type_df[nc] = None

            super().bulk_upsert(type_df)

    @check_types
    @guard("edit")
    def bulk_delete(self, df: DataFrame[RemoveRunMetaEntryFrameSchema]) -> None:
        self.check_df_access(df)
        return super().bulk_delete(df)



from typing import ClassVar

import pandas as pd
from sqlalchemy.exc import NoResultFound

from ixmp4 import db
from ixmp4.data import types
from ixmp4.data.abstract import iamc as abstract
from ixmp4.data.auth.decorators import guard

from ..unit import Unit
from . import base
from .variable import Variable


class Measurand(base.BaseModel):
    NotFound: ClassVar = abstract.Measurand.NotFound
    NotUnique: ClassVar = abstract.Measurand.NotUnique
    DeletionPrevented: ClassVar = abstract.Measurand.DeletionPrevented

    __table_args__ = (db.UniqueConstraint("variable__id", "unit__id"),)

    variable__id: types.Integer = db.Column(
        db.Integer, db.ForeignKey("iamc_variable.id"), nullable=False, index=True
    )
    variable: types.Mapped[Variable] = db.relationship(
        "Variable",
        backref="measurands",
        foreign_keys=[variable__id],
        lazy="select",
    )

    unit__id: types.Integer = db.Column(
        db.Integer, db.ForeignKey("unit.id"), nullable=False, index=True
    )
    unit: types.Mapped[Unit] = db.relationship(
        "Unit",
        backref="measurands",
        foreign_keys=[unit__id],
        lazy="select",
    )

    created_at = db.Column(db.DateTime)
    created_by = db.Column(db.String(255))


class MeasurandRepository(
    base.Creator[Measurand],
    base.Retriever[Measurand],
    base.Enumerator[Measurand],
    abstract.MeasurandRepository,
):
    model_class = Measurand

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

    @guard("view")
    def get(self, variable_name: str, unit__id: int) -> Measurand:
        exc: db.sql.Select = (
            db.select(Measurand)
            .join(Measurand.variable)
            .where(Measurand.unit__id == unit__id)
            .where(Variable.name == variable_name)
        )

        try:
            return self.session.execute(exc).scalar_one()
        except NoResultFound:
            raise Measurand.NotFound

    def add(self, variable_name: str, unit__id: int) -> Measurand:
        q_exc: db.sql.Select = db.select(Variable).where(Variable.name == variable_name)
        try:
            variable = self.session.execute(q_exc).scalar_one()
        except NoResultFound:
            variable = Variable(name=variable_name)

        measurand = Measurand(variable=variable, unit__id=unit__id)
        self.session.add(measurand)
        self.session.commit()
        return measurand

    @guard("edit")
    def create(self, *args, **kwargs) -> Measurand:
        return super().create(*args, **kwargs)

    def select(self, *args, **kwargs) -> db.sql.Select:
        return db.select(Measurand)

    @guard("view")
    def list(self, *args, **kwargs) -> list[Measurand]:
        return super().list(*args, **kwargs)

    @guard("view")
    def tabulate(self, *args, **kwargs) -> pd.DataFrame:
        return super().tabulate(*args, **kwargs)



from functools import partial

import typer

echo = typer.echo
good = partial(typer.secho, fg=typer.colors.GREEN)
bad = partial(typer.secho, fg=typer.colors.RED)
error = partial(typer.secho, fg=typer.colors.RED, err=True)
info = partial(typer.secho, fg=typer.colors.CYAN)
important = partial(typer.secho, fg=typer.colors.MAGENTA)
secondary = partial(typer.secho, fg=typer.colors.BRIGHT_BLACK)



from ixmp4.data.db import filters as base
from ixmp4.data.db.run import Run
from ixmp4.db import filters, utils

from .model import IndexSet


class RunFilter(base.RunFilter, metaclass=filters.FilterMeta):
    def join(self, exc, **kwargs):
        if not utils.is_joined(exc, Run):
            exc = exc.join(Run, onclause=IndexSet.run__id == Run.id)
        return exc


class OptimizationIndexSetFilter(
    base.OptimizationIndexSetFilter, metaclass=filters.FilterMeta
):
    def join(self, exc, session=None):
        return exc



from enum import Enum
from typing import ClassVar, Protocol

import pandas as pd
from pydantic import StrictBool, StrictFloat, StrictInt, StrictStr

from ixmp4.data import types

from . import base

# as long as all of these are `Strict` the order does not matter
StrictMetaValue = StrictBool | StrictInt | StrictFloat | StrictStr
MetaValue = bool | int | float | str


class RunMetaEntry(base.BaseModel, Protocol):
    """Run meta entry model."""

    class Type(str, Enum):
        INT = "INT"
        STR = "STR"
        FLOAT = "FLOAT"
        BOOL = "BOOL"

        @classmethod
        def from_pytype(cls, type_):
            return RunMetaEntry._type_map[type_]

    run__id: types.Integer
    "Foreign unique integer id of a run."
    key: types.String
    "Key for the entry. Unique for each `run__id`."
    type: types.String
    "Datatype of the entry's value."
    value: types.Integer | types.Float | types.Integer | types.Boolean
    "Value of the entry."

    _type_map: ClassVar[dict] = {
        int: Type.INT,
        str: Type.STR,
        float: Type.FLOAT,
        bool: Type.BOOL,
    }

    def __str__(self) -> str:
        return f"<RunMetaEntry {self.id} run__id={self.run__id} \
            key={self.key} value={self.value}"


class RunMetaEntryRepository(
    base.Creator,
    base.Retriever,
    base.Deleter,
    base.Enumerator,
    base.BulkUpserter,
    base.BulkDeleter,
    Protocol,
):
    def create(
        self,
        run__id: int,
        key: str,
        value: MetaValue,
    ) -> RunMetaEntry:
        """Creates a meta indicator entry for a run.

        Parameters
        ----------
        run__id : str
            The unique integer id of a run.
        key : str
            The key of the entry.
        value : str, int, bool or float
            The value of the entry.

        Raises
        ------
        :class:`ixmp4.core.exceptions.RunMetaEntryNotUnique`:
            If the entry with `run__id` and `key` is not unique.

        Returns
        -------
        :class:`ixmp4.data.base.RunMetaEntry`:
            The created entry.
        """
        ...

    def get(self, run__id: int, key: str) -> RunMetaEntry:
        """Retrieves a run's meta indicator entry.

        Parameters
        ----------
        run__id : str
            The unique integer id of a run.
        key : str
            The key of the entry.

        Raises
        ------
        :class:`ixmp4.data.abstract.RunMetaEntry.NotFound`:
            If the entry with `run__id` and `key` does not exist.

        Returns
        -------
        :class:`ixmp4.data.base.Run`:
            The retrieved entry.
        """
        ...

    def delete(self, id: int) -> None:
        """Deletes a run's meta indicator entry.
        Warning: No recovery of deleted data shall be possible via ixmp
        after the execution of this function.

        Parameters
        ----------
        id : int
            The unique integer ids of entries to delete.

        Raises
        ------
        :class:`ixmp4.data.abstract.RunMetaEntry.NotFound`:
            If the entry with `id` does not exist.
        """
        ...

    def list(
        self,
        **kwargs,
    ) -> list[RunMetaEntry]:
        r"""Lists run's meta indicator entries by specified criteria.

        Parameters
        ----------
        \*\*kwargs: any
            Filter parameters as specified in
            `ixmp4.data.db.meta.filter.RunMetaEntryFilter`.

        Returns
        -------
        Iterable[:class:`ixmp4.data.abstract.RunMetaEntry`]:
            List of run meta indicator entries.
        """
        ...

    def tabulate(
        self,
        join_run_index: bool = False,
        **kwargs,
    ) -> pd.DataFrame:
        r"""Tabulates run's meta indicator entries by specified criteria.

        Parameters
        ----------
        \*\*kwargs: any
            Filter parameters as specified in
            `ixmp4.data.db.meta.filter.RunMetaEntryFilter`.

        Returns
        -------
        :class:`pandas.DataFrame`:
            A data frame with the columns:
                - id
                - run__id
                - key
                - type
                - value
        """
        ...

    def bulk_upsert(self, df: pd.DataFrame) -> None:
        """Upserts a dataframe of run meta indicator entries.

        Parameters
        ----------
        df : :class:`pandas.DataFrame`
            A data frame with the columns:
                - run__id
                - key
                - value
                - type
        """
        ...

    def bulk_delete(self, df: pd.DataFrame) -> None:
        """Deletes run meta indicator entries as specified per dataframe.
        Warning: No recovery of deleted data shall be possible via ixmp
        after the execution of this function.

        Parameters
        ----------
        df : :class:`pandas.DataFrame`
            A data frame with the columns:
                - run__id
                - key

        """
        ...



from typing import List, Protocol

import pandas as pd

from ixmp4.data import types

from .. import base
from ..docs import DocsRepository


class IndexSet(base.BaseModel, Protocol):
    """IndexSet data model."""

    name: types.String
    """Unique name of the IndexSet."""
    run__id: types.Integer
    """The id of the :class:`ixmp4.data.abstract.Run` for which this IndexSet is
    defined. """

    elements: types.JsonList
    """Unique list of str or int."""

    created_at: types.DateTime
    "Creation date/time. TODO"
    created_by: types.String
    "Creator. TODO"

    def __str__(self) -> str:
        return f"<IndexSet {self.id} name={self.name}>"


class IndexSetRepository(
    base.Creator,
    base.Retriever,
    base.Enumerator,
    Protocol,
):
    docs: DocsRepository

    def create(self, run_id: int, name: str) -> IndexSet:
        """Creates an IndexSet.

        Parameters
        ----------
        run_id : int
            The id of the :class:`ixmp4.data.abstract.Run` for which this IndexSet is
            defined.
        name : str
            The name of the IndexSet.

        Raises
        ------
        :class:`ixmp4.data.abstract.IndexSet.NotUnique`:
            If the IndexSet with `name` already exists.

        Returns
        -------
        :class:`ixmp4.data.abstract.IndexSet`:
            The created IndexSet.
        """
        ...

    def get(self, run_id: int, name: str) -> IndexSet:
        """Retrieves an IndexSet.

        Parameters
        ----------
        run_id : int
            The id of the :class:`ixmp4.data.abstract.Run` for which this IndexSet is
            defined.
        name : str
            The unique name of the IndexSet.

        Raises
        ------
        :class:`ixmp4.data.abstract.IndexSet.NotFound`:
            If the IndexSet with `name` does not exist.

        Returns
        -------
        :class:`ixmp4.data.abstract.IndexSet`:
            The retrieved IndexSet.
        """
        ...

    def list(self, *, name: str | None = None, **kwargs) -> list[IndexSet]:
        r"""Lists IndexSets by specified criteria.

        Parameters
        ----------
        name : str
            The name of an IndexSet. If supplied only one result will be returned.
        # TODO: Update kwargs
        \*\*kwargs: any
            More filter parameters as specified in
            `ixmp4.data.db.iamc.variable.filters.VariableFilter`.

        Returns
        -------
        Iterable[:class:`ixmp4.data.abstract.IndexSet`]:
            List of IndexSets.
        """
        ...

    def tabulate(self, *, name: str | None = None, **kwargs) -> pd.DataFrame:
        r"""Tabulate IndexSets by specified criteria.

        Parameters
        ----------
        name : str
            The name of an IndexSet. If supplied only one result will be returned.
        # TODO: Update kwargs
        \*\*kwargs: any
            More filter parameters as specified in
            `ixmp4.data.db.iamc.variable.filters.VariableFilter`.

        Returns
        -------
        :class:`pandas.DataFrame`:
            A data frame with the columns:
                - id
                - name
                - elements
        """
        ...

    def add_elements(
        self, indexset_id: int, elements: int | List[int | str] | str
    ) -> None:
        """Adds elements to an existing IndexSet.

        Parameters
        ----------
        indexset_id : int
            The id of the target IndexSet.
        elements : int | List[int | str] | str
            The elements to be added to the IndexSet.

        Returns
        -------
        None:
            Due to compatibility with ixmp.
        """
        ...



"""

Run the web api with:

.. code:: bash

   ixmp4 server start [--host 127.0.0.1] [--port 8000]

This will start ixmp4’s asgi server. Check
``http://127.0.0.1:8000/v1/<platform>/docs``.

"""

from fastapi import FastAPI

from .rest import v1

app = FastAPI()

app.mount("/v1/{platform}", v1)



from datetime import datetime
from typing import ClassVar

import pandas as pd
from pydantic import Field

from ixmp4.data import abstract

from .. import base
from ..docs import Docs, DocsRepository


class Variable(base.BaseModel):
    NotFound: ClassVar = abstract.Variable.NotFound
    NotUnique: ClassVar = abstract.Variable.NotUnique
    DeletionPrevented: ClassVar = abstract.Variable.DeletionPrevented

    id: int
    name: str

    created_at: datetime | None = Field(None)
    created_by: str | None = Field(None)


class VariableDocsRepository(DocsRepository):
    model_class = Docs
    prefix = "docs/iamc/variables/"


class VariableRepository(
    base.Creator[Variable],
    base.Retriever[Variable],
    base.Enumerator[Variable],
    abstract.VariableRepository,
):
    model_class = Variable
    prefix = "iamc/variables/"

    def __init__(self, client, *args, **kwargs) -> None:
        super().__init__(client, *args, **kwargs)
        self.docs = VariableDocsRepository(self.client)

    def create(
        self,
        name: str,
    ) -> Variable:
        return super().create(name=name)

    def get(self, name: str) -> Variable:
        return super().get(name=name)

    def enumerate(self, **kwargs) -> list[Variable] | pd.DataFrame:
        return super().enumerate(**kwargs)

    def list(self, **kwargs) -> list[Variable]:
        return super()._list(json=kwargs)

    def tabulate(self, **kwargs) -> pd.DataFrame:
        return super()._tabulate(json=kwargs)



from ixmp4.data.db import filters as base
from ixmp4.db import filters


class TimeSeriesFilter(base.TimeSeriesFilter, metaclass=filters.FilterMeta):
    id: filters.Id
    run: base.RunFilter = filters.Field(default=base.RunFilter())
    region: base.RegionFilter | None
    variable: base.VariableFilter | None
    unit: base.UnitFilter | None

    def join(self, exc, **kwargs):
        return exc
